var net = require('net'),
    events = require('events'),
    sys = require('sys'),
    mpstream = require ('./msgpack-stream');

var REQUEST  = 0;
var RESPONSE = 1;
var NOTIFY   = 2;
var MAX_SEQID = Math.pow(2,32)-1;

function RPCResponse (stream, seqid) {
  this.stream = stream;
  this.seqid  = seqid;
}

RPCResponse.prototype.result = function(args) {
  this.stream.respond(this.seqid, null, args);
}

RPCResponse.prototype.error = function(error) {
  this.stream.respond(this.seqid, error, null);
}

// The heart of the beast, used for both server and client
var id = 0;
var MsgpackRPCStream = function(stream, handler) {

    events.EventEmitter.call(this);
    var self              = this;
    this.last_seqid       = undefined;
    this.stream           = stream;
    this.handler          = handler;
    this.cbs              = [];
    this.timeout          = undefined;
    this.id               = id++; 
    this.eofcb            = undefined;
    
    this.msgpack_stream = new mpstream.Stream(this.stream);

    //-------------------------------------------------------------

    // 
    // This stream is a packetized input stream, so we only get full
    // messages.  This callback is called when a new message arrives..
    //
    this.msgpack_stream.on('msg', function(msg) {
	var type = msg.shift();
	switch(type) { 
	case REQUEST:
            var seqid  = msg[0];
            var method = msg[1];
            var params = msg[2];
            var response = new RPCResponse(self, seqid); 
	    
            self.invokeHandler(method, params.concat(response));
            self.emit('request', method, params, response);
            break;
	case RESPONSE:
            var seqid  = msg[0];
            var error  = msg[1];
            var result = msg[2];
	    
            if(self.cbs[seqid]) {
		self.triggerCb(seqid, [error, result]);
            } else {
		var err = new Error("unexpected response with " +
				    "unrecognized seqid (" + seqid + ")" );
		self.emit('error', err);
            }
            break;
	case NOTIFY:
            var method = msg[0];
            var params = msg[1];
	    
            self.invokeHandler(method, params);
            self.emit('notify', method, params);
            break;
	}
    });

    //-------------------------------------------------------------

    this.stream.on('connect', function() { 
	self.emit('ready'); 
    });

    //-------------------------------------------------------------

    this.stream.on('end', function() { 
	
	self.stream.end(); 
	
	// For all of those people still waiting for a reply, it's now
	// hopelss.
	self.failCbs(new Error("connection closed by peer")); 
	
	if (self.eofcb) { self.eofcb (); }
    });
    
    //-------------------------------------------------------------

    this.stream.on('timeout', function() { 
	self.failCbs(new Error("connection timeout")); 
    });

    //-------------------------------------------------------------

    this.stream.on('close', function(had_error) {
	if(had_error) return; 
	self.failCbs(new Error("connection closed locally"));
    });

    //-------------------------------------------------------------

    this.stream.on('error', function(error) { 
	self.failCbs(error); 
    });
}

//=======================================================================

sys.inherits(MsgpackRPCStream, events.EventEmitter);

MsgpackRPCStream.prototype.triggerCb = function(seqid, args) {
    this.cbs[seqid].apply(this, args);
    delete this.cbs[seqid];
}

//-----------------------------------------------------------------------

MsgpackRPCStream.prototype.setErrorHandlers = function () {
    var self = this;
    this.stream.on ('error', function (err) { self.failCbs (err); });
};

//-----------------------------------------------------------------------

MsgpackRPCStream.prototype.failCbs = function(error) {
    for(var seqid in this.cbs) { 
	this.triggerCb(seqid, [error]) 
    }
}

//-----------------------------------------------------------------------

MsgpackRPCStream.prototype.invokeHandler = function(method, params) {
    if(this.handler) {
	if(this.handler[method]) {
	    this.handler[method].apply(this.handler, params);
	} else {
	    response.error(new Error("unknown method")); 
	}
    }
}

//-----------------------------------------------------------------------

MsgpackRPCStream.prototype.nextSeqId = function() {
    if(this.last_seqid == undefined) {
	return this.last_seqid = 0;
    } else if(this.last_seqid > MAX_SEQID ) {
	return this.last_seqid = 0;
    } else {
	return this.last_seqid += 1;
    }
}

//-----------------------------------------------------------------------

// End of the stream, not necessarily a failure....
MsgpackRPCStream.prototype.setEofCb = function (c) { this.eofcb = c; }

//-----------------------------------------------------------------------

MsgpackRPCStream.prototype.invoke = function() {
    var self   = this;
    var seqid  = this.nextSeqId();
    var method = arguments[0];
    var cb     = arguments[arguments.length - 1];
    var args = [];

    for(var i = 1;i < arguments.length - 1;i++) {
	args.push(arguments[i]);
    }
    
    this.cbs[seqid] = cb;
    if(this.timeout) {
	setTimeout(
	    function() { if(self.cbs[seqid]) 
		self.triggerCb(seqid, ["timeout"]); }, 
	    this.timeout
	);
    }
    if(this.stream.writable) { 
	return this.msgpack_stream.send([REQUEST, seqid, method, args]) 
    };
}

//-----------------------------------------------------------------------

MsgpackRPCStream.prototype.respond = function(seqid, error, result) {
    if(this.stream.writable) { 
	return this.msgpack_stream.send([RESPONSE, seqid, error, result]) 
    };
}

//-----------------------------------------------------------------------

MsgpackRPCStream.prototype.notify = function(method, params) {
    var method = arguments[0];
    var args   = [];
    for(var i = 1;i < arguments.length;i++) {
	args.push(arguments[i]);
    }
    
    if(this.stream.writable) { 
	return this.msgpack_stream.send([NOTIFY, method, args]) 
    };
}

//-----------------------------------------------------------------------

MsgpackRPCStream.prototype.setTimeout = function(timeout) {
    this.timeout = timeout;
}

//-----------------------------------------------------------------------

MsgpackRPCStream.prototype.close = function() {
    this.stream.end();
}

//=======================================================================

var Transport = function (hostname, port) {
    
    if (!hostname || hostname.length == 0 || hostname == "-") {
	hostname = null;
    }

    this.hostname = hostname;
    this.port = port;
    this._connect_cb = null;
    this.tcp_stream = null;
    this.rpc_stream = null;

    //-----------------------------------------

    this._trigger = function (ok, msg) {

	if (msg) {
	    console.log ("shit! error = " + msg);
	}
	
	if (this._connect_cb) {
	    var cb = this._connect_cb;
	    this._connect_cb = null;
	    if (ok) {
		this.rpc_stream.setErrorHandlers ();
	    }
	    cb (ok, msg);
	}
    };

    //-----------------------------------------

    this.connect = function (cb) {
	var self = this;
	this._connect_cb = cb;
	this.tcp_stream = new net.createConnection (port, hostname);
	this.rpc_stream = new MsgpackRPCStream (this.tcp_stream);
	this.rpc_stream.on ('ready', 
			    function () { self._trigger (true, null); });
	this.tcp_stream.on ('error', 
			    function (err) { self._trigger (false, err); });
    };

    //-----------------------------------------

    this.notify = function () {
	this.rpc_stream.notify.apply (this.rpc_stream, arguments);
    };

    //-----------------------------------------

    this.invoke = function () {
	this.rpc_stream.invoke.apply (this.rpc_stream, arguments);
    };

    //-----------------------------------------

    this.close = function () {
	this.tcp_stream.end ();
    };

    //-----------------------------------------

    return this;
};

//=======================================================================

exports.Transport = Transport;
exports.createTransport = function (hostname, port) {
    return new Transport (hostname, port);
};

//=======================================================================

function Client (transport, program) {
    this.transport = transport;
    this.program = program;

    //--------------------------------------------------

    this.invoke = function () {
	if (this.program) {
	    arguments[0] = this.program + "." + arguments[0];
	}
	this.transport.invoke.apply (this.transport, arguments);
    };

    //--------------------------------------------------

    this.notify = function () {
	if (this.program) {
	    arguments[0] = this.program + "." + arguments[0];
	}
	this.transport.notify.apply (this.transport, arguments);
    };

    //--------------------------------------------------

    return this;
};

//=======================================================================

exports.Client = Client;

//=======================================================================

exports.createClient = function (hostname, port, cb) {

    if (hostname && (!hostname.length || hostname == "-")) {
	hostname = null;
    }

    var conn = new net.createConnection (port, hostname);
    var s = new MsgpackRPCStream (conn);
    if (cb) { s.on ('ready', cb); }
    return s;
};

//=======================================================================

var Server = function (listener) {
    net.Server.call(this);
    var self = this;
    this.handler = undefined;
    this.verbose = false;
    this.name = "";
    
    this.on('connection', function(tcp_stream) {
	if (this.verbose) {
	    var s = "";
	    if (this.name) { s = this.name + ": "; }
	    s += ("new connection from: " + stream.remoteAddress + ":" 
		  + stream.remotePort);
	    console.log (s);
	    
	}
	tcp_stream.on('end', function() { tcp_stream.end(); });
	var rpc_stream = new MsgpackRPCStream(tcp_stream, self.handler);
	if (listener) {
	    listener(rpc_stream);
	}
    });
}

//-----------------------------------------------------------------------

sys.inherits(Server, net.Server);

Server.prototype.setHandler = function(handler) {
    this.handler = handler;
}

//-----------------------------------------------------------------------

Server.prototype.setVerbose = function (v, n) {
    this.verbobse = v;
    this.name = n;
};

//-----------------------------------------------------------------------

exports.createServer = function(listener) {
  return new Server(listener);
}

//=======================================================================

var SessionPool = exports.SessionPool = function() {
    this.clients = {};
};

//-----------------------------------------------------------------------

SessionPool.prototype.getClient = function(port, hostname) {
    var address = hostname + ":" + port;
    if(this.clients[address]) {
	return this.clients[address];
    } else {
	return this.clients[address] = exports.createClient(port, hostname);
    }
};

//-----------------------------------------------------------------------

SessionPool.prototype.closeClients = function() {
    for(var i in this.clients) this.clients[i].stream.end();
};

//=======================================================================

